/* this file implements the classes for bfile */
#include <math.h>
#include <string.h>
#include <memory.h>
#include "../blockfile/cache.h"
#include "bfile.h"
//-----------------------------------------------------
BEntry::BEntry()
{
	bitchars=NULL;
}

//-----------------------------------------------------

BEntry::~BEntry()
{
	delete []bitchars;
}
//-----------------------------------------------------

void BEntry::del_son()
{	
	if (son_ptr) 
	{ 
		delete son_ptr; son_ptr = NULL; 
	}
}

//-----------------------------------------------------

int BEntry::get_size()
{   
	return sizeof(tm) + sizeof(son) + bitcharlen;
}

//-----------------------------------------------------

BNode *BEntry::get_son()
{
	if (son_ptr == NULL)
		son_ptr = new BNode(my_rtree, son);
	return son_ptr;
}

//-----------------------------------------------------

void BEntry::init_entry(RTree *_my_rtree)
{  
   tm=-1; son=-1; son_ptr = NULL; 
   my_rtree = _my_rtree;
   hashnum=my_rtree->hashnum;
   bitlen=my_rtree->bitlen;
   bitmaplen=my_rtree->bitmaplen;  
   bitcharlen=my_rtree->bitcharlen;  
   bitchars=new unsigned char[bitcharlen];
}

//-----------------------------------------------------

void BEntry::read_from_buffer(char *_buf)
{
	int i = 0;
	memcpy(&son, _buf+i, sizeof(son));
	i+=sizeof(son);
	memcpy(&tm, _buf + i, sizeof(tm)); 
	i+=sizeof(tm);
	memcpy(bitchars, _buf+i, bitcharlen); 
	i+=bitcharlen;
}
//-----------------------------------------------------

void BEntry::write_to_buffer(char *_buf)
{
	int i = 0;
	memcpy(_buf+i, &son, sizeof(son));
	i+=sizeof(son);
	memcpy(_buf+i, &tm, sizeof(tm));
	i+=sizeof(tm);
	memcpy(_buf + i, bitchars, bitcharlen); 
	i+=bitcharlen;
}

/*****************************************************************
this function gets the bitmaps of an entry, by converting its bitchar array to a bool array
coded by yufei tao, may 2003
******************************************************************/

bool * BEntry::get_bitmaps()
{
	bool *bitmaps=new bool[hashnum*bitlen];
	char2bitmap(bitmaps, bitchars, hashnum*bitlen);
	return bitmaps;
}

/*****************************************************************
this function updates the bitmap of this entry by "or-ing" with the input bitmap
para:
bitchar: the input bitmap
REMINDER: SET THE NODE'S DIRTY BIT ON!
coded by yufei tao, may 2003
******************************************************************/

void BEntry::bitmap_or(unsigned char *_bitchars)
{
	bool *ibitmaps=new bool[bitmaplen], *bitmaps=new bool[bitmaplen];
	char2bitmap(ibitmaps, _bitchars, bitmaplen);
	char2bitmap(bitmaps, bitchars, bitmaplen);
	for (int i=0; i<bitmaplen; i++)
		bitmaps[i]|=ibitmaps[i];
	bitmap2char(bitmaps, bitchars, bitmaplen, bitcharlen);
	delete []bitmaps;
	delete []ibitmaps;
}

/*****************************************************************
this function updates the information of this entry according to its child node
para:
succ: its child node
REMINDER: SET THE NODE'S DIRTY BIT ON!
coded by yufei tao, may 2003
******************************************************************/

void BEntry::set_to_son(BNode *_succ)
{
	son=_succ->block;
	son_ptr=_succ;
	tm=_succ->entries[0].tm;

	bool *bitmaps=_succ->get_bitmaps();
//	char2bitmap(bitmaps, bitchars, bitmaplen);
	bitmap2char(bitmaps, bitchars, bitmaplen, bitcharlen);
	delete []bitmaps;
}

/*****************************************************************
copy the content of the input entry to the current entry
coded by yufei tao, may 2003
******************************************************************/

void BEntry::copy(BEntry *_e)
{
	bitcharlen=_e->bitcharlen;
	memcpy(bitchars, _e->bitchars, bitcharlen);
	bitlen=_e->bitlen;
	bitmaplen=_e->bitmaplen;
	hashnum=_e->hashnum;
	my_rtree=_e->my_rtree;
	son=_e->son;
	son_ptr=_e->son_ptr;
	tm=_e->tm;
}

/*****************************************************************
this constructor creates a brandnew node from the disk
para:
my_rtree: point to the host tree
coded by yufei tao, may 2003
******************************************************************/

BNode::BNode(RTree *_my_rtree)
{
	//=====initialize the common atrributes=====
	init_block(_my_rtree);
	//==========================================

	num_entries = 0;

	//=====get a block from the disk=====
	int file_len = my_rtree -> file -> get_blocklength();
	char *temp_buf = new char[file_len];
	block = my_rtree -> file -> append_block(temp_buf);
	delete [] temp_buf;
	//===================================
	dirty = true;
}

/*****************************************************************
this constructor restores a node from the disk
para:
my_rtree: point to the host tree
blk: the disk address
coded by yufei tao, may 2003
******************************************************************/

BNode::BNode(RTree *_my_rtree, int _blk)
{
	//=====initialize the common atrributes=====
	init_block(_my_rtree);
	//==========================================

	//=====get a block from the disk=====
	block=_blk;
	int file_len = my_rtree->file->get_blocklength();
	char *temp_buf = new char[file_len];
	if (my_rtree->cache) 
		my_rtree->cache->read_block(temp_buf, _blk, my_rtree);
	else 
		my_rtree->file->read_block(temp_buf, _blk);
	
	read_from_buffer(temp_buf);
	delete [] temp_buf;
	//===================================

	dirty =false;
}
//-----------------------------------------------------

BNode::~BNode()
{	
	if (dirty)
	{
		char *buf = new char[my_rtree -> file -> get_blocklength()];
		write_to_buffer(buf);
		if (my_rtree -> cache)
			my_rtree -> cache -> write_block(buf, block, my_rtree);
		else
			my_rtree -> file -> write_block(buf, block);
		delete [] buf;
	}
	delete [] entries; 
}

/*****************************************************************
//this function adds a new entry in this node as the parent entry of the given node _new_nd
coded by yufei tao, may 2003
******************************************************************/

void BNode::add_new_child(BNode *_new_nd)
{
	entries[num_entries].set_to_son(_new_nd);
	entries[num_entries].del_son();
	num_entries ++; 
	dirty = true;
}

/*****************************************************************
this function gets the bitmaps of a node, which corresponds to the union of the bitmaps of its entries
coded by yufei tao, may 2003
******************************************************************/

bool * BNode::get_bitmaps()
{
	bool *bitmaps=new bool[bitlen*hashnum];
	for (int i=0; i<bitlen*hashnum; i++)
		bitmaps[i]=false;
	for (i=0; i<num_entries; i++)
	{
		bool *ebitmaps=entries[i].get_bitmaps();
		for (int j=0; j<bitlen*hashnum; j++)
			bitmaps[j]|=ebitmaps[j];
		delete []ebitmaps;
	}
	return bitmaps;
}

/*****************************************************************
this function enters the information of a new leaf entry
para:
tm: the time when the insertion is made. note that the time must be the current time. namely, all the
insertions must be made chronologically
bitchar: 
coded by yufei tao, may 2003
******************************************************************/

void BNode::add_leaf(int _tm, unsigned char *_bitchars)  
{
	if (num_entries>0 && entries[num_entries-1].tm==_tm)
	{
		entries[num_entries-1].bitmap_or(_bitchars);
	}
	else
	{
		bool same=true;
		if (num_entries>0)
		{
			for (int i=0; i<bitcharlen; i++)
				if (_bitchars[i]!=entries[num_entries-1].bitchars[i])
					same=false;
		}
		else
			same=false;
		if (!same)
		{
			entries[num_entries].tm=_tm;
			memcpy(entries[num_entries].bitchars, _bitchars, bitcharlen);
			num_entries++;
		}
	}
	
	dirty = true;	
}

//-----------------------------------------------------

bool BNode::chk_ovrflw()
{  
	if (num_entries==node_capacity)
		return true;
	else 
		return false;
}

//-----------------------------------------------------

int BNode::choose_subtree(int _key)
{  
	//it is always the last branch be selected becase the keys are inserted
	//in chronological order.
	return num_entries - 1;
}

//-----------------------------------------------------

int BNode::header_size()
{   
	return sizeof(level) + sizeof(num_entries); 
}

//-----------------------------------------------------

void BNode::init_block(RTree *_my_rtree)
{
	my_rtree = _my_rtree;

	int file_len = my_rtree -> file -> get_blocklength();
	BEntry *tmpe=new BEntry();
	tmpe->init_entry(my_rtree);
	page_capacity = (file_len - header_size()) / tmpe->get_size();
	delete tmpe;

	entries = new BEntry[page_capacity];
	for (int i = 0; i < page_capacity; i ++)
		entries[i].init_entry(my_rtree);

	node_capacity = page_capacity - 2;
	hashnum=my_rtree->hashnum;
	bitlen=my_rtree->bitlen;
	bitmaplen=my_rtree->bitmaplen;
	bitcharlen=my_rtree->bitcharlen;
}

/*****************************************************************
this function inserts a new entry into the subtree of this node
para:
tm: the time when the insertion is made. note that the time must be the current time. namely, all the
insertions must be made chronologically
bitchar: 
new_nd: needed in case of overflow
coded by yufei tao, may 2003
******************************************************************/

BFile_Insrt BNode::insert(int _tm, unsigned char *_bitchar, BNode **_new_nd)
{
	BFile_Insrt ret = BF_NONE;
	if (level > 0)
    {    
		int follow = choose_subtree(_tm);
		BNode *new_nd = NULL;
		BNode *succ = entries[follow].get_son();
		BFile_Insrt c_ret = succ -> insert(_tm, _bitchar, &new_nd);
		
		//update the bitmap of the parent entry (or-ing it with bitchar)
		entries[follow].set_to_son(succ);
		entries[follow].del_son();

		if (c_ret == BF_SPLIT)
		{
			add_new_child(new_nd);  
			if (chk_ovrflw())
				ret = trt_ovrflw(_new_nd); 
		}
		dirty = true;
		return ret;
	}

	//leaf nodes here
	add_leaf(_tm, _bitchar);
	if (chk_ovrflw())
		ret = trt_ovrflw(_new_nd);
	dirty = true;
	return ret;
}

//-----------------------------------------------------

void BNode::read_from_buffer(char *_buf)
{
	int i = 0;
	memcpy(&level, _buf + i, sizeof(level));
	i += sizeof(level);
	memcpy(&num_entries, _buf + i, sizeof(num_entries)); 
	i += sizeof(num_entries);
	
	int esize=entries[0].get_size();
	for (int j = 0; j < num_entries; j ++)
	{  
		entries[j].read_from_buffer(_buf + i); 
	    i+=esize; 
	}
}

/*****************************************************************
handles the overflow of a node. since the insertions are made chronologically, an overflow is treated by
creating a new node with only one entry.
para:
tm: the time when the insertion is made. note that the time must be the current time. namely, all the
insertions must be made chronologically
bitchar: 
new_nd: needed in case of overflow
coded by yufei tao, may 2003
******************************************************************/

BFile_Insrt BNode::trt_ovrflw(BNode **_new_nd)
{
	//error("bnode overflows... this coding path has not been tested yet...\n", true);
	(*_new_nd) = new BNode(my_rtree);
	BNode *new_nd = *_new_nd;
	new_nd->level=level;
	new_nd->entries[0].copy(&entries[num_entries-1]);
	num_entries--;
	new_nd->num_entries=1;
	return BF_SPLIT;
}

//-----------------------------------------------------

void BNode::write_to_buffer(char *_buf)
{
	int i = 0;
	memcpy(_buf + i, &level, sizeof(level)); 
	i += sizeof(level);
	memcpy(_buf + i, &num_entries, sizeof(num_entries)); 
	i += sizeof(num_entries);
	
	int esize=entries[0].get_size();
	for (int j = 0; j < num_entries; j ++)
	{  
		entries[j].write_to_buffer(_buf + i); 
		i += esize;
	}
}

/*****************************************************************
this function performs an aggregate query using the arb-tree
para:
qbitmaps: the accumuated qbitmaps so far
qmbr: the query spatial mbr
qt: the query interval
et: the last timestamp in the lifespan of this node (passed from the parent entry)
coded by yufei tao, may 2003
******************************************************************/

void BNode::agg_qry(bool *_qbitmaps, float *_qmbr, float *_qt, int _ed)
{
//debugging---
//fprintf(my_rtree->qfp, "%d\n", block);
//------------
	if (level==0)
	{
		for (int i=0; i<num_entries; i++)
		{
			int st=entries[i].tm, ed;
			if (i!=num_entries-1)
				ed=entries[i+1].tm;
			else
				ed=_ed;
			if (max(st,_qt[0])<=min(ed-1,_qt[1]))
			{
				bool *bitmaps=new bool[bitmaplen];
				char2bitmap(bitmaps, entries[i].bitchars, bitmaplen);
				for (int j=0; j<bitmaplen; j++)
					_qbitmaps[j]|=bitmaps[j];
				delete []bitmaps;
			}
		}
		return;
	}

	for (int i=0; i<num_entries; i++)
	{
		int st=entries[i].tm, ed;
		if (i!=num_entries-1)
			ed=entries[i+1].tm;
		else
			ed=_ed;

		if (_qt[0]<=st && ed-1<=_qt[1])
		{
			bool *bitmaps=new bool[bitmaplen];
			char2bitmap(bitmaps, entries[i].bitchars, bitmaplen);
			for (int j=0; j<bitmaplen; j++)
				_qbitmaps[j]|=bitmaps[j];
			delete []bitmaps;
		}
		else if (max(st,_qt[0])<=min(ed-1,_qt[1]))
		{
			BNode *succ=entries[i].get_son();
			succ->agg_qry(_qbitmaps, _qmbr, _qt, ed);
			entries[i].del_son();
		}
	}
}

/*****************************************************************
this function gets the bitmap at a particular timestamp
para:
tm: the requested timestamp
bitchar: the bitchar retruned
coded by yufei tao, may 2003
******************************************************************/

void BNode::get_reg_bitchar(int _tm, unsigned char *_bitchars)
{
	if (level==0)
	{
		int pos=0;
		for (int i=0; i<num_entries; i++)
		{
			if (i==num_entries-1 || (entries[i].tm<=_tm && entries[i+1].tm>_tm))
			{	
				pos=i; break;
			}
		}
		memcpy(_bitchars, entries[pos].bitchars, bitcharlen);
		return;
	}

	for (int i=0; i<num_entries; i++)
	{
		int pos=0;
		for (int j=0; j<num_entries; j++)
		{
			if (j==num_entries-1 || (entries[j].tm<=_tm && entries[j+1].tm>_tm))
			{	
				pos=j; break;
			}
		}
		BNode *succ=entries[pos].get_son();
		succ->get_reg_bitchar(_tm, _bitchars);
		bool *succbitmaps=succ->get_bitmaps();
		bool *thisbitmaps=new bool[bitmaplen];
		char2bitmap(thisbitmaps, entries[pos].bitchars, bitmaplen);
		for (j=0; j<bitmaplen; j++)
			if (succbitmaps[j]!=thisbitmaps[j])
				error("found a bug\n", true);
		delete []thisbitmaps;
		delete []succbitmaps; 
		entries[pos].del_son();
	}
}

/*****************************************************************
use this function to init a new b-tree
para:
_my_rtree: the host r-tree
coded by yufei tao, may 2003
******************************************************************/

BTree::BTree(RTree *_my_rtree)
{
	my_rtree=_my_rtree;
	root_ptr=new BNode(my_rtree);
	root_ptr->level=0;
	root=root_ptr->block;
	del_root();
}

/*****************************************************************
use this function to restore a b-tree from the disk
para:
_my_rtree: the host r-tree
root: the root of the tree
coded by yufei tao, may 2003
******************************************************************/

BTree::BTree(RTree *_my_rtree, int _root)
{
	my_rtree=_my_rtree;
	root=_root;
	root_ptr=NULL;
}

//-----------------------------------------------------

BTree::~BTree()
{
	del_root();
}
//-----------------------------------------------------

void BTree::del_root()
{
	if (root_ptr)
	{  delete root_ptr; root_ptr = NULL; }
}

/*****************************************************************
this function inserts a new entry into the this tree
para:
tm: the time when the insertion is made. note that the time must be the current time. namely, all the
insertions must be made chronologically
bitchar: 
this function returns true if the root has changed
coded by yufei tao, may 2003
******************************************************************/

bool BTree::insert(int _tm, unsigned char *_bitchar)
{
	bool ret=false;
	load_root();
	
	BNode *new_nd = NULL;
	BFile_Insrt c_ret = root_ptr -> insert(_tm, _bitchar, &new_nd);

	if (c_ret == BF_SPLIT)
	{
		BNode *new_root_ptr=new BNode(my_rtree);
		new_root_ptr->level=root_ptr->level+1;
		new_root_ptr->add_new_child(root_ptr);
		new_root_ptr->add_new_child(new_nd);
		root = new_root_ptr -> block; root_ptr = new_root_ptr;
		ret=true;
	}

	del_root();
	return ret;
}

//-----------------------------------------------------

void BTree::load_root()
{
	if (!root_ptr) root_ptr=new BNode(my_rtree, root);
}

/*****************************************************************
this function performs an aggregate query using the arb-tree
para:
qbitmaps: the accumuated qbitmaps so far
qmbr: the query spatial mbr
qt: the query interval
ed: please give the last timestamp of the history
coded by yufei tao, may 2003
******************************************************************/

void BTree::agg_qry(bool *_qbitmaps, float *_qmbr, float *_qt, float _ed)
{
	load_root();
	root_ptr->agg_qry(_qbitmaps, _qmbr, _qt, _ed);
	del_root();
}

/*****************************************************************
this function gets the bitmap at a particular timestamp
para:
tm: the requested timestamp
bitchar: the bitchar retruned
coded by yufei tao, may 2003
******************************************************************/

void BTree::get_reg_bitchars(int _tm, unsigned char *_bitchars)
{
	load_root();
	root_ptr->get_reg_bitchar(_tm, _bitchars);
	del_root();
}

/*****************************************************************
this function loads the heap (for query) with the qualifying entries
in the roots of the B-trees
para:
qbitmaps: the initial result
qmbr: the spatial query mbr
qt: the query interval
ed:	please give the last timestamp of the history
coded by yufei tao, june 2003
******************************************************************/

void BTree::load_heap(bool *_qbitmaps, float *_qmbr, float *_qt, float _ed)
{
	load_root();
//debugging---
//fprintf(my_rtree->qfp, "%d\n", root_ptr->block);
//------------
	HeapEntry *he=new HeapEntry();
	he->init_HeapEntry(my_rtree->bitcharlen);
	int num_entries=root_ptr->num_entries;
	for (int i=0; i<num_entries; i++)
	{
		int st=root_ptr->entries[i].tm, ed;
		if (i!=num_entries-1)
			ed=root_ptr->entries[i+1].tm;
		else
			ed=_ed;

		if (root_ptr->level==0)
		{
			if (max(st,_qt[0])<=min(ed-1,_qt[1]))
			{
				bool *bitmaps=new bool[my_rtree->bitmaplen];
				char2bitmap(bitmaps, root_ptr->entries[i].bitchars, my_rtree->bitmaplen);
				for (int j=0; j<my_rtree->bitmaplen; j++)
					_qbitmaps[j]|=bitmaps[j];
				delete []bitmaps;	
			}
		}
		else
		{
			if (_qt[0]<=st && ed-1<=_qt[1])
			{
				bool *bitmaps=new bool[my_rtree->bitmaplen];
				char2bitmap(bitmaps, root_ptr->entries[i].bitchars, my_rtree->bitmaplen);
				for (int j=0; j<my_rtree->bitmaplen; j++)
					_qbitmaps[j]|=bitmaps[j];
				delete []bitmaps;
			}
			else if (max(st,_qt[0])<=min(ed-1,_qt[1]))
				{
					//init a heap entry now
					memcpy(he->bitchars, root_ptr->entries[i].bitchars, my_rtree->bitcharlen);
					he->level=root_ptr->level;
					he->son=root_ptr->entries[i].son;
					he->key=get_one_number(he->bitchars, my_rtree->bitmaplen);
					he->ed=ed;
					my_rtree->hp->insert(he);
				}
		}
	}
	delete he;
	del_root();
}
