/*****************************************************************
this package contains the implementation of paper
"title goes here"
Functions included are for the purposes of data, signature file generation,
accuracy test, etc.
for questions, comments and bug reports please contact yufei tao at 
taoyf@cs.ust.hk. in case i change my affiliation please search
"yufei tao" on google to reach my current homepage
may 2003
*****************************************************************/

#include <math.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include <fstream.h>
#include "rand.h"
#include "./rand/random.h"
#include "./rtree/rtree.h"
#include "./blockfile/cache.h"
#define min(a, b) (((a) < (b))? (a) : (b)  )
#define max(a, b) (((a) > (b))? (a) : (b)  )

float new_uniform(int _d_num)
{
	float base=1;
	float sum=0; 
	for (int i=0; i<_d_num; i++)
	{
		int digit=uniform(0, 10);
		if (digit==10) digit=9;
		sum+=base*digit;
		base*=10;
	}
	return sum;
}

float new_uniform(float _min, float _max)
{
	float ran_base=9999999;
	float ran=new_uniform(7);
	return ran/ran_base*(_max-_min)+_min;
}

void fatal(char *_msg)
{
	printf(_msg);
	exit(1);
}

/*****************************************************************
this function writes an integer into a file opened in the binary mode
Coded by yufei tao may 2003
*****************************************************************/
void binary_int_write(FILE *_fp, int _n)
{
	char *c=(char *)&_n;
	for(int i=0; i<sizeof(int); i++)
	{
		fputc(*c, _fp); c++;
	}
}

/*****************************************************************
this function reads an integer From a file opened in the binary mode
Coded by yufei tao may 2003
*****************************************************************/
void binary_int_read(FILE *_fp, int *_n)
{
	char *c=(char *)_n;
	for(int i=0; i<sizeof(int); i++)
	{
		*c=fgetc(_fp); c++;
	}
}

/*****************************************************************
this function transforms a bitmap to a char array suitable for storage
bitmap: the bitmap to be converted
charmap: the char array to be returned
len: the number of bits in bitmap
if all bits are false then this funciton return true;
Coded by yufei tao may 2003
*****************************************************************/
bool bitmap2char(bool *_bitmap, unsigned char *_charmap, int _len, int _clen)
{
	bool allfalse=true;
	for (int i=0; i<_clen; i++)
		_charmap[i]=0;
	for (i=0; i<_len; i++)
	{
		if (_bitmap[i])
		{
			allfalse=false;
			int sub=i/8, offset=7-i%8;
			unsigned char tmp=1;
			tmp=tmp<<offset;
			_charmap[sub]|=tmp;
		}
	}
	return allfalse;
}

/*****************************************************************
this function is the reverse of bitmap2char
charmap: the char array to be returned
len: the number of bits in bitmap
clen: the number elements in charmap
if all bits are false then this funciton return true;
Coded by yufei tao may 2003
*****************************************************************/
void char2bitmap(bool *_bitmap, unsigned char *_charmap, int _len)
{
	for (int i=0; i<_len; i++)
		_bitmap[i]=false;
	for (i=0; i<_len; i++)
	{
		int sub=i/8, offset=7-i%8;
		if ((_charmap[sub]>>offset)%2)
			_bitmap[i]=true;
	}
}

/*****************************************************************
this funciton normalizes a 2D dataset into a specified range
NOTE****: vary the input file format in the code, before using this
function
para:
fname: input file name
ofname: output file name
d: the dimensionality
U: the intended size of the universe 
format: the format of the data file
Coded by Jimeng Sun 29/06/02
*****************************************************************/

void normalize(char* fname, char* ofname, int _d, double _U, int _format)
{
	FILE *ifp, *ofp;
	ifp=fopen(fname, "r");
	ofp=fopen(ofname, "w");
	int tmp;
	float *x=new float[2*_d];
	double *xmin=new double[_d];
	double *xmax=new double[_d];

	for (int i=0; i<_d; i++)
	{
		xmin[i]=1e20; xmax[i]=-1e20;
	}
	
	int cnt=0;
	
	printf("file %s\n",fname);
	if(!ifp)
	{
		cout<<"file not exists!\n";
		exit(0);
	}
	int format=_format;
	while(!feof(ifp))
	{
		switch(format)
		{
		case 1:
			fscanf(ifp, "%d", &tmp);
			for (i=0; i<_d; i++)
				fscanf(ifp, " %f %f", &x[i], &x[i+_d]);
			fscanf(ifp, "\n");
			break;
		case 2:
			for (i=0; i<_d; i++)
				fscanf(ifp, " %f %f", &x[2*i], &x[2*i+1]);
			fscanf(ifp, "\n");
			break;
		case 3:
			for (i=0; i<_d; i++)
			{
				fscanf(ifp, " %f", &x[2*i]);
				x[2*i+1]=x[2*i];
			}
			fscanf(ifp, "\n");
			break;
		case 4:
			fscanf(ifp, "%d", &tmp);
			for (i=0; i<_d; i++)
				fscanf(ifp, " %f %f", &x[2*i], &x[2*i+1]);
			fscanf(ifp, "\n");
			break;
		case 5:
			for (i=0; i<_d; i++)
				fscanf(ifp, " %f %f", &x[i], &x[i+_d]);
			fscanf(ifp, "\n");
			break;
		}

		for (i=0; i<_d; i++)
		{
			if (x[2*i]>x[2*i+1])
			{
				printf("data inconsistency captured\n"); exit(1);
			}
//			x[2*i+1]=x[2*i];
			if (x[2*i]<xmin[i])
				xmin[i]=x[2*i];
			if (x[2*i+1]>xmax[i])
				xmax[i]=x[2*i+1];
			if (xmin[i]>xmax[i])
			{
				printf("error: xmin is larger than xmax\n");
				exit(1);
			}
		}
		
		if(cnt%5000 == 0)
			printf("read %d records\n", cnt);

		cnt++;
	}

	double *xlen=new double[_d];
	for (i=0; i<_d; i++)
	{
		xlen[i]=xmax[i]-xmin[i];
		printf("on dim %d: min=%.3f, max=%.3f\n", i+1, xmin[i], xmax[i]);
	}
	printf("read %d records\n", cnt);
	
	fclose(ifp);
	ifp=fopen(fname, "r");
	cnt=0;
	while(!feof(ifp))
	{
		switch(format)
		{
		case 1:
			fscanf(ifp, "%d", &tmp);
			for (i=0; i<_d; i++)
				fscanf(ifp, " %f %f", &x[i], &x[i+_d]);
			fscanf(ifp, "\n");
			break;
		case 2:
			for (i=0; i<_d; i++)
				fscanf(ifp, " %f %f", &x[2*i], &x[2*i+1]);
			fscanf(ifp, "\n");
			tmp=cnt+1;
			break;
		case 3:
			for (i=0; i<_d; i++)
			{
				fscanf(ifp, " %f", &x[2*i]);
				x[2*i+1]=x[2*i];
			}
			fscanf(ifp, "\n");
			tmp=cnt+1;
			break;
		case 4:
			fscanf(ifp, "%d", &tmp);
			for (i=0; i<_d; i++)
				fscanf(ifp, " %f %f", &x[2*i], &x[2*i+1]);
			fscanf(ifp, "\n");
			break;
		case 5:
			for (i=0; i<_d; i++)
				fscanf(ifp, " %f %f", &x[i], &x[i+_d]);
			fscanf(ifp, "\n");
			break;
		}
		for (i=0; i<_d; i++)
		{
			x[2*i]=(x[2*i]-xmin[i])/xlen[i]*_U; 
			x[2*i+1]=(x[2*i+1]-xmin[i])/xlen[i]*_U; 
//			x[2*i+1]=x[2*i];
			if (x[2*i]<0 || x[2*i+1]>_U)
			{
				printf("exceeding the universe %.3f, %.3f\n", x[2*i], x[2*i+1]);
				exit(1);
			}
		}
		
		if(cnt%5000==0)
			printf("finish %d records\n",cnt);
		
		fprintf(ofp, "%d", cnt);
		for (i=0; i<_d; i++)
			fprintf(ofp, " %f %f", x[2*i], x[2*i+1]);
		fprintf(ofp, "\n");
		cnt++;
	}

	delete []xlen;
	delete []xmin; delete []xmax;
	delete []x;

	fclose(ifp);
	fclose(ofp);
}

/*****************************************************************
this function scales the y-axis of the dataset so that the triangles
are equilateral
ofname: the output file name
d: dimensionality
U: the universe axis
N: the cardinality
distri: the distribution you want (1-uniform, 2-zipf)
coded by yufei tao Dec 27 2002
*****************************************************************/

void gen_simple_ds(char *_ofname, int _d, float _U, int _N, int _distri)
{
	FILE *ofp=fopen(_ofname, "w");
	if (!ofp)
	{
		printf("could not open or create files\n"); exit(1);
	}
	for (int i=0; i<_N; i++)
	{
		int id; 
		fprintf(ofp, "%d ", i);
		for (int j=0; j<_d; j++)
		{
			float x;
			switch (_distri)
			{
			case 1:
				x=new_uniform(0, _U);
				break;
			case 2:
				bool again=true;
				while (again)
				{
					again=false;
					x=zipf(0, _U, 0.8);
					if (x<0 || x>_U)
						again=true;
				}
				break;
			}
			fprintf(ofp, "%f %f ", x, x);
//			fprintf(ofp, "%f ", x);
		}
		fprintf(ofp, "\n");
	}
		
	fclose(ofp);
}

/*****************************************************************
this function generates datasets as follows. we take a spatial dataset, specified with the name dsname.
the centroids of its rectangles are used as the airports. next the locations 
of N airplanes are created from these airports. Each airplane is associated with a destination and a
velocity. After reaching the destination, the airplane randomly selects the next destination and repeats
this process again until the time expires. At each timestamp we record for each airport the ids of the
airplanes that have it as the nearest neighbor

dsname: the file name of the spatial dataset
ofname: the output file name
rnum: the number of regions in dsname
anum: the number of airports
N: number of moving objects
T: number of evolving timestamps
vmin:
vmax: the velocity range

coded by yufei tao, may 2003
*****************************************************************/

struct airplane
{
	float startpos[2];
	float endpos[2];
	float dist;  //distance between the start and end
	float vel; 
	float vx, vy;  //velocities on the x- and y-dimensions respectively
	float starttm;  //start time
	float endtm; //end time
	int ssub, esub;
	int year; //weight of this airplane, specifying the number years it has served
};

void gen_moving_ds(char *_dsname, char *_tname, char *_ofname, int _rnum, int _N, int _T,
				   float _vmin, float _vmax)
{
	//inner tuning variables -------------------------------------
	int MAX_SERVED_YEARS=10; 
	//------------------------------------------------------------

	FILE *fp=fopen(_dsname, "r");
	if (!fp)
		fatal("could not open the spatial ds\n");
	FILE *ofp=fopen(_ofname, "w");
	if (!ofp)
		fatal("could not create the output file\n");
	RTree *rt=new RTree(_tname, NULL);

	float *regions=new float[4*_rnum]; 
	int cnt=0;
	printf("reading the spatial dataset...\n");
	while (!feof(fp))
	{
		if (cnt==_rnum)
			fatal("the spatial dataset is larger than you thought...\n");
		int id;
		fscanf(fp, "%d ", &id);
		if (id!=cnt)
			error("the spatial file is not sorted according to id\n", true);
		for (int i=0; i<4; i++)
			fscanf(fp, "%f", &regions[4*cnt+i]);
		fscanf(fp, "\n");
		cnt++;
	}
	printf("\tread %d regions\n", cnt);
	if (cnt<_rnum)
	{
		printf("\twarning: the dataset is smaller than you thought\n");
		_rnum=cnt;
	}
	
	printf("taking the centroids...\n");
	float *centroids=new float[2*_rnum];
	for (int i=0; i<_rnum; i++)
	{
		centroids[2*i]=(regions[4*i]+regions[4*i+1])/2;
		centroids[2*i+1]=(regions[4*i+2]+regions[4*i+3])/2;
	}

	printf("generating the inital information of the airplanes...\n");
	airplane *ap=new airplane[_N];
	for (i=0; i<_N; i++)
	{
		int ssub=uniform(0, _rnum);
		if (ssub==_rnum)
			ssub=_rnum-1;
		ap[i].startpos[0]=centroids[2*ssub];
		ap[i].startpos[1]=centroids[2*ssub+1];
		int esub=uniform(0, _rnum);
		if (esub==_rnum)
			esub=_rnum-1;
		ap[i].endpos[0]=centroids[2*esub];
		ap[i].endpos[1]=centroids[2*esub+1];
		float distx=ap[i].endpos[0]-ap[i].startpos[0], disty=ap[i].endpos[1]-ap[i].startpos[1];
		ap[i].dist=sqrt(pow(distx, 2)+pow(disty, 2));
		ap[i].vel=uniform(_vmin, _vmax);
		ap[i].vx=ap[i].vel*distx/ap[i].dist;
		ap[i].vy=ap[i].vel*disty/ap[i].dist;
		ap[i].starttm=0; 
		ap[i].endtm=ap[i].dist/ap[i].vel;
		ap[i].year=(int)zipf(1, MAX_SERVED_YEARS, 0.8); 
ap[i].ssub=ssub; ap[i].esub=esub;
	}

	printf("now start generating data for subsequent timestamps...\n");
	for (i=1; i<=_T; i++)
	{
		printf("\ttimestamp %d...\n", i);

		for (int j=0; j<_N; j++)
		{
			//compute the current location of this airplane...
			float pos[2];
			pos[0]=ap[j].startpos[0]+ap[j].vx*(i-ap[j].starttm);
			pos[1]=ap[j].startpos[1]+ap[j].vy*(i-ap[j].starttm);
			
			//find the nearest region
			int reg=-1;
			rt->nn(pos, reg); 

			fprintf(ofp, "%d %d %d %d\n", i, reg, j, ap[j].year);
//			for (int k=0; k<4; k++)
//				fprintf(ofp, "%f ", regions[4*reg+k]);
		}

		for (j=0; j<_N; j++)
		{
			if (i>=ap[j].endtm)
			{
				ap[j].startpos[0]=ap[j].endpos[0];
				ap[j].startpos[1]=ap[j].endpos[1];
				int esub=uniform(0, _rnum);
				if (esub==_rnum)
					esub=_rnum-1;
				ap[j].endpos[0]=centroids[2*esub];
				ap[j].endpos[1]=centroids[2*esub+1];
				float distx=ap[j].endpos[0]-ap[j].startpos[0], disty=ap[j].endpos[1]-ap[j].startpos[1];
				ap[j].dist=sqrt(pow(distx, 2)+pow(disty, 2));
				ap[j].vel=uniform(_vmin, _vmax);
				ap[j].vx=ap[j].vel*distx/ap[j].dist;
				ap[j].vy=ap[j].vel*disty/ap[j].dist;
				ap[j].starttm=i; 
				ap[j].endtm=i+ap[j].dist/ap[j].vel;
			}
		}
	}

	delete []ap;
	delete []centroids;
	delete []regions;

	delete rt;
	fclose(ofp);
	fclose(fp);
}

/*****************************************************************
this function generates the signatures for the spatio-temporal dataset

dsname: the file name of the spatial dataset
ofname: the output file name
rnum: the number of regions in dsname
T: number of evolving timestamps
hashnum: the number of hash functions used
bitlen: the number of bits in each signature (pls set it as a multiple of 8)
pcsa: toggle if pcsa is applied
wt: toggle if generating signatures on sum 
coded by yufei tao, may 2003
*****************************************************************/

//auxiliary function: this function sets the bitmaps according to the eid given, for count sketches
void set_count_bitmap(bool *_bitmaps, int _bitlen, int _eid, int _hashnum, bool _pcsa)
{
	Random ran(_eid);  //the same set of random values will be produced if the seeds are the same
	int ranint=ran.nextInt();
	Random ran2(ranint);
	if (_pcsa)
	{
		int bitMapChoice = ran2.nextInt(_hashnum); //choose which signature to set the bits
		
		int x=0;
		//after the following while loop x=0 with probability 0.5, x=1 with prob 0.25, etc...
		while (ran2.nextBoolean())
			x++;
		if (x>_bitlen)
		{
			printf("x=%d\n", x);
			fatal("\tError: x>bitlen in gen_singnatures\n");
		}
		_bitmaps[bitMapChoice*_bitlen+x]=true;
	}
	else
	{
		for (int i=0; i<_hashnum; i++)
		{
			int x=0;
			while (ran2.nextBoolean())
				x++;
			if (x>_bitlen)
			{
				printf("x=%d\n", x);
				fatal("\tError: x>bitlen in gen_singnatures\n");
			}
			_bitmaps[i*_bitlen+x]=true;
		}
	}
}

//translation of kachitvichyanukul88binomial BINV code
int binomial_pick(int nIn, double pIn,Random rg)
{
	double q = 1.0 - pIn;
	double s = pIn / q;
	double a = ((double)(nIn+1))*s;

	double r = pow(q,(double)nIn);
	double u = rg.nextDouble();
	int x = 0;
	while (u>r)
	{
		u -= r;
		++x;
		r*= (a/x) -s;
	}
	return x;
}

//auxiliary function: this function sets the bitmaps according to the eid, value given, for count sketches
void set_sum_bitmap(bool *_bitmaps, int _bitlen, int _eid, int _value, int _hashnum, bool _pcsa)
{
	Random rg(_eid + _value);
	//add individual values so bulk adds are equally sized
	int tmpv=_value;
	for (; (tmpv % _hashnum) !=0; --tmpv)
	{
		int bitMapChoice = rg.nextInt(_hashnum);

		int x =0;
		while (rg.nextBoolean())
		{
			x++;
		}
		_bitmaps[bitMapChoice*_bitlen+x] = true;
	}
	int bulkAddSize = _value / _hashnum;

	int fill_length=0; //the number of bits that are set 1
	//now decide fill_length
	if (_value>10)
	{
		double value = (double) _value;
		double log_value = log(value);
		double tmp = log (value / (log_value*log_value));
		fill_length = (int) floor(tmp/log(2.0));   
	}

	for (int i =0; i<_hashnum;i++)
	{
		for (int j=0;j<fill_length;j++)
			_bitmaps[_bitlen*i+j] = true;	

		int tmp = binomial_pick(bulkAddSize,pow(0.5,(double)fill_length),rg);
		int remaining = ((fill_length ==0)?bulkAddSize:tmp);
		for (j=0;j<remaining;j++)
		{
			int x = fill_length;
			while (rg.nextBoolean())
			{
				x++;
			}
			_bitmaps[i*_bitlen+x] = true;
		}
	}
}

void gen_signatures(char *_sfname, char *_ofname, int _hashnum, int _bitlen, bool _pcsa, bool _wt)
{
	if (_bitlen%8)
		fatal("bitlen must be a multiple of 8\n");
	FILE *fp=fopen(_sfname, "r");
	if (!fp)
		fatal("could not open the input file\n");
	FILE *ofp=fopen(_ofname, "wb");
	if (!ofp)
		fatal("could not create the output file\n");

	printf("computing the sigantures...\n");
	bool *bitmaps=new bool[_bitlen*_hashnum];
	for (int i=0; i<_bitlen*_hashnum; i++)
		bitmaps[i]=false;

	int tm, rid, eid, weight;
	int last_tm=-1, last_rid;
	bool first=true;
	int bitmapcharlen=_hashnum*_bitlen/8;
	unsigned char *bitmapchar=new unsigned char [bitmapcharlen];
	while (!feof(fp))
	{
		fscanf(fp, "%d %d %d %d\n", &tm, &rid, &eid, &weight);
		if (tm>last_tm)
			printf("timestamp %d...\n", tm);
		if (first)
		{
			first=false; last_tm=tm; last_rid=rid;
		}
		else
		{
			if (tm<last_tm || (tm==last_tm && rid<last_rid))
				fatal("the file is not sorted\n");
		}
		
		if (tm!=last_tm || rid!=last_rid)
		{
			//write the bitmap information to the disk
			bool allfalse=bitmap2char(bitmaps, bitmapchar, _hashnum*_bitlen, bitmapcharlen);
			if (!allfalse)
			{
				binary_int_write(ofp, last_tm);
				binary_int_write(ofp, last_rid);
//				fprintf(ofp, "%d %d ", last_tm, last_rid);
				for (int j=0; j<bitmapcharlen; j++)
//					fprintf(ofp, "%c ", bitmapchar[j]);
					fputc(bitmapchar[j], ofp);
			}
			
			for (i=0; i<_hashnum*_bitlen; i++)
				bitmaps[i]=false;
			last_tm=tm; last_rid=rid;
		}
		
		if (_wt)
			set_sum_bitmap(bitmaps, _bitlen, eid, weight, _hashnum, _pcsa);
		else
			set_count_bitmap(bitmaps, _bitlen, eid, _hashnum, _pcsa);
	}

	//write the bitmap information of the last region to the disk
	bool allfalse=bitmap2char(bitmaps, bitmapchar, _hashnum*_bitlen, bitmapcharlen);
	if (!allfalse)
	{
		binary_int_write(ofp, last_tm);
		binary_int_write(ofp, last_rid);
//				fprintf(ofp, "%d %d ", last_tm, last_rid);
		for (int j=0; j<bitmapcharlen; j++)
//				fprintf(ofp, "%c ", bitmapchar[j]);
			fputc(bitmapchar[j], ofp);
	}
	delete []bitmaps;
	binary_int_write(ofp, -1);  //signifiying the end of the file
	fclose(ofp);
	fclose(fp);
}

/*****************************************************************
this function performs an external sort on the region-id matching file
produced by gen_moving_ds
para:
sfname: source file name
tfname: target file name
psize: how many records in each partition of the source file
dim: the dimensionality of the data space
coded by yufei tao, may 2003
******************************************************************/

struct item
{
	int tm;
	int rid;
	int eid;
	int year;
};

int item_compare(const void *_p1, const void *_p2)
{
	item *p1=(item *)_p1, *p2=(item *)_p2;

	if (p1->tm<p2->tm)
		return -1;
	else if (p1->tm>p2->tm)
		return 1;

	if (p1->rid<p2->rid)
		return -1;
	else
		return 1;

	if (p1->eid<p2->eid)
		return -1;
	else
		return 1;

	return 0;
}

typedef FILE * fileptr;

void ext_sort(char *_sfname, char *_tfname, int _psize)
{
	FILE *fp=fopen(_sfname, "r");
	if (!fp)
		fatal("sorry i could not open the source file\n");
	FILE *ofp=fopen(_tfname, "w");
	if (!ofp)
		fatal("sorry i could not create the target file\n");
	//first init an array with _psize elements
	item *element=new item[_psize];
	//partition phase
	printf("reading the source file...\n");
	int cnt=0, fcnt=0, tcnt=0;
	while (!feof(fp))
	{
		fscanf(fp, "%d %d %d %d\n", 
			&(element[cnt].tm), &(element[cnt].rid), &(element[cnt].eid), &(element[cnt].year));
		cnt++;
		tcnt++;

		if (cnt==_psize)
		{
			qsort(element, cnt, sizeof(item), item_compare); 

			fcnt++;
			char fname[50];
			itoa(fcnt, fname, 10); 
			FILE *mfp=fopen(fname, "w");
			if (!mfp)
				fatal("sorry i could not create the intermediate file\n");
			printf("writing intermediate file # %d...", fcnt);
			for (int i=0; i<cnt; i++)
				fprintf(mfp, "%d %d %d %d\n", element[i].tm, element[i].rid, element[i].eid,
				element[i].year);
			fclose(mfp);
			printf("done\n");
			cnt=0;
		}
	}
	if (cnt<_psize && cnt!=0)
	{
		qsort(element, cnt, sizeof(item), item_compare);

		fcnt++;
		char fname[50];
		itoa(fcnt, fname, 10); 
		FILE *mfp=fopen(fname, "w");
		if (!mfp)
			fatal("sorry i could not create the intermediate file\n");
		printf("writing intermediate file # %d...", fcnt);
		for (int i=0; i<cnt; i++)
			fprintf(mfp, "%d %d %d %d\n", element[i].tm, element[i].rid, element[i].eid,
			element[i].year);
		fclose(mfp);
		printf("done\n");
		cnt=0;
	}

	delete []element;
	fclose (fp);

	printf("read totally %d records\n", tcnt);
	
	//the merge phase
	printf("starting the merging phase\n");
	FILE **mfp=new fileptr[fcnt];
	for (int i=0; i<fcnt; i++)
	{
		char fname[50];
		itoa(i+1, fname, 10);
		mfp[i]=fopen(fname, "r");
		if (!mfp[i])
			fatal("sorry i could not open the intermediate file...\n");
	}

	int endcnt=0; //how many files have reached the end
	item *fele=new item[fcnt]; 
	//get the first record of each file
	for (i=0; i<fcnt; i++)
	{
		if (!feof(mfp[i]))
			fscanf(mfp[i], "%d %d %d %d\n", &(fele[i].tm), &(fele[i].rid), &(fele[i].eid), &(fele[i].year));
		else
		{
			endcnt++;
			fele[i].tm=1000000;
		}
	}
	int mtcnt=0;
	while (endcnt<fcnt)
	{
		//output the smallest element in fele
		int smallpt=0;
		for (int i=1; i<fcnt; i++)
		{
			if (item_compare(&(fele[i]),&(fele[smallpt]))==-1)
				smallpt=i;
		}

		fprintf(ofp, "%d %d %d %d\n", fele[smallpt].tm, fele[smallpt].rid, fele[smallpt].eid,
			fele[smallpt].year);
		mtcnt++;
		if (mtcnt%100000==0)
			printf("merged %d records (%.3f%)\n", mtcnt, (float)mtcnt/tcnt*100);

		//get the next record for the corresponding file
		if (!feof(mfp[smallpt]))
			fscanf(mfp[smallpt], "%d %d %d %d\n",
			&(fele[smallpt].tm), &(fele[smallpt].rid), &(fele[smallpt].eid), &(fele[smallpt].year));
		else
		{
			endcnt++;
			fele[smallpt].tm=1000000;
		}
	}
	delete []fele;

	for (i=0; i<fcnt; i++)
	{
		fclose(mfp[i]);
		char fname[50];
		itoa(i+1, fname, 10);
		remove(fname);
	}
	delete []mfp;

	fclose(ofp);
}

/*****************************************************************
this function tests the accuracy of the signature technique, usiong the
signature file and the spatio-temporal dataset generated by gen_moving_ds
para:
sigfname: signature file name
stfname: spatio-temporal file name
qrid: the range of the region ids queried
qt: the range of query timestamps included
N: the number of possible eids
hashnum: the number of hash functions used
bitlen: number of bits per signature
coded by yufei tao, may 2003
******************************************************************/

void test_accuracy(char *_sigfname, char *_stfname, int *_qrid, float *_qt, int _N, int _hashnum, int _bitlen)
{
	if (_bitlen%8)
		fatal("bitlen must be a multiple of 8\n");
	//tuning variables -------------------------------------------
	int maxperregion=100;
	//ends -------------------------------------------------------

	FILE *sfp=fopen(_sigfname, "rb"); //s stands for signature
	if (!sfp)
		fatal("could not open the signature file\n");
	FILE *rfp=fopen(_stfname, "r"); //r stands for raw
	if (!rfp)
		fatal("could not open the raw data file\n");

	int qridlen=_qrid[1]-_qrid[0]+1;
	//first scan the signature file to get the estimate
	printf("scanning the signature file...\n");
	int last_tm=-1;
	int bitmapcharlen=_hashnum*_bitlen/8; 
	unsigned char *bitmapchar=new unsigned char[bitmapcharlen];
	bool *bitmaps=new bool[_hashnum*_bitlen];
	for (int i=0; i<_hashnum*_bitlen; i++)
		bitmaps[i]=false;
	bool *thisbitmaps=new bool[_hashnum*_bitlen];
	int tm,rid;
	binary_int_read(sfp, &tm);
	while (tm!=-1 && last_tm<=_qt[1])
	{
		binary_int_read(sfp, &rid);

		for (int i=0; i<bitmapcharlen; i++)
			bitmapchar[i]=fgetc(sfp);

		if (_qrid[0]<=rid && rid<=_qrid[1] && _qt[0]<=tm && tm<=_qt[1])
		{
			for (int i=0; i<_hashnum*_bitlen; i++)
				thisbitmaps[i]=false;
			char2bitmap(thisbitmaps, bitmapchar, _hashnum*_bitlen);
			for (i=0; i<_hashnum*_bitlen; i++)
				if (thisbitmaps[i]) 
					bitmaps[i]=true;
		}
		last_tm=tm;
		binary_int_read(sfp, &tm);
	}
	//now perform estimate
	int total=0;
	for (i=0; i<_hashnum; i++)
	{
		int pos=0;
		for (int j=0; j<_bitlen; j++)
			if (!bitmaps[i*_bitlen+j])
			{
				pos=j; break;
			}
		total+=pos;
	}
	float nmaps=(float)_hashnum;
	float tmp=pow(2.0, ((float)total)/nmaps);
	float est = (nmaps/0.77351)*tmp;
	printf("\test=%f\n", est);

	delete []thisbitmaps;
	delete []bitmaps;
	delete []bitmapchar;

	//first scan the raw file to get the actual
	printf("scanning the raw file...\n");
	bool *eids=new bool[_N];
	for (i=0; i<_N; i++)
		eids[i]=false;
	int num=0;
	
	last_tm=-1;
	while (!feof(rfp) && last_tm<=_qt[1]) 
	{
		int tm,rid,eid;
		fscanf(rfp, "%d %d %d", &tm, &rid, &eid);

		if (_qrid[0]<=rid && rid<=_qrid[1] && _qt[0]<=tm && tm<=_qt[1])
		{
			if (!eids[eid])
			{
				eids[eid]=true; num++;
			}
		}
		
		last_tm=tm;
	}
	printf("\tact=%d\n", num);
	delete []eids;

	fclose(rfp);
	fclose(sfp);
}

/*****************************************************************
this function tests the accuracy of the signature technique, usiong the
signature file and the spatio-temporal dataset generated by gen_moving_ds
the difference from the previous version is that it takes a query mbr to 
filter the regions instead of simply by their ids
para:
dfname: the spatial data file
sigfname: signature file name
stfname: spatio-temporal file name
qmbr: the query mbr
qt: the range of query timestamps included
N: the number of possible eids
hashnum: the number of hash functions used
bitlen: number of bits per signature
pcsa: whether pcsa is used
coded by yufei tao, may 2003
******************************************************************/

void test_accuracy(char *_dfname, char *_sigfname, char *_stfname, float *_qmbr, float *_qt, int _rnum, int _N,
				   int _hashnum, int _bitlen, bool _pcsa)
{
	if (_bitlen%8)
		fatal("bitlen must be a multiple of 8\n");

	FILE *dfp=fopen(_dfname, "r");
	if (!dfp)
		fatal("could not open the spatial data file\n");
	FILE *sfp=fopen(_sigfname, "rb"); //s stands for signature
	if (!sfp)
		fatal("could not open the signature file\n");
	FILE *rfp=fopen(_stfname, "r"); //r stands for raw
	if (!rfp)
		fatal("could not open the raw data file\n");

	printf("reading the spatial file...\n");
	float *reg=new float[4*_rnum];
	int cnt=0;
	while (!feof(dfp))
	{
		if (cnt==_rnum)
			error("\tthe spatial dataset's cardinality is larger than you thought...\n", true);
		int id; 
		fscanf(dfp, "%d ", &id);
		if (id!=cnt)
			error("the spatial file is not sorted according to id\n", true);
		for (int i=0; i<4; i++)
			fscanf(dfp, "%f ", &reg[4*cnt+i]);
		fscanf(dfp, "\n");
		cnt++;
	}
	printf("\tread %d regions\n", cnt);
	if (cnt!=_rnum)
		printf("\twarning: the spatial dataset's cardinality is smaller than you thought...\n");

	//first scan the signature file to get the estimate
	printf("scanning the signature file...\n");
	int last_tm=-1;
	int bitmapcharlen=_hashnum*_bitlen/8; 
	unsigned char *bitmapchar=new unsigned char[bitmapcharlen];
	bool *bitmaps=new bool[_hashnum*_bitlen];
	for (int i=0; i<_hashnum*_bitlen; i++)
		bitmaps[i]=false;
	bool *thisbitmaps=new bool[_hashnum*_bitlen];
	int tm,rid;
	binary_int_read(sfp, &tm);
	while (tm!=-1 && last_tm<=_qt[1])
	{
		binary_int_read(sfp, &rid);

		for (int i=0; i<bitmapcharlen; i++)
			bitmapchar[i]=fgetc(sfp);

		float *tmp=overlapRect(2, _qmbr, &reg[rid*4]);
		if (tmp)
		{
			delete []tmp;
			if (_qt[0]<=tm && tm<=_qt[1])
			{
				for (int i=0; i<_hashnum*_bitlen; i++)
					thisbitmaps[i]=false;
				char2bitmap(thisbitmaps, bitmapchar, _hashnum*_bitlen);
				for (i=0; i<_hashnum*_bitlen; i++)
					if (thisbitmaps[i]) 
						bitmaps[i]=true;
			}
		}
		last_tm=tm;
		binary_int_read(sfp, &tm);
	}
	//now perform estimate
	
	int total=0;
	for (i=0; i<_hashnum; i++)
	{
		int pos=0;
		for (int j=0; j<_bitlen; j++)
			if (!bitmaps[i*_bitlen+j])
			{
				pos=j; break;
			}
		total+=pos;
	}
	float nmaps=(float)_hashnum;
	float tmp=pow(2.0, ((float)total)/nmaps);
	float est;
	if (_pcsa)
		est=(nmaps/0.77351)*tmp;
	else
		est=tmp/0.77351;
	printf("\test=%f\n", est);

	delete []thisbitmaps;
	delete []bitmaps;
	delete []bitmapchar;

	//first scan the raw file to get the actual
	printf("scanning the raw file...\n");
	bool *eids=new bool[_N];
	for (i=0; i<_N; i++)
		eids[i]=false;
	int num=0;
	
	last_tm=-1;
	while (!feof(rfp) && last_tm<=_qt[1]) 
	{
		int tm,rid,eid, year;
		fscanf(rfp, "%d %d %d %d", &tm, &rid, &eid, &year);

		float *tmp=overlapRect(2, _qmbr, &reg[rid*4]);
		if (tmp)
		{
			delete []tmp;
			if (_qt[0]<=tm && tm<=_qt[1])
			{
				if (!eids[eid])
				{
					eids[eid]=true; num++;
				}
			}
		}
		last_tm=tm;
	}
	printf("\tact=%d\n", num);
	delete []eids;
	delete []reg;

	fclose(dfp);
	fclose(rfp);
	fclose(sfp);
}

/*****************************************************************
this function generates a query workload with the same query parameters
each tuple in the query workload has the format <x0,x1,y0,y1,t0,t1,res>
where res is the actual result
para:
qfname: the query file name
qs: spatial length
qt: temporal length
U: the universe axis length [0, U]
T: the number of timestamps in history [1,T]
qnum: number of queries in the workload
N: number of data regions
dfname: the spatial data file
stfname: the spatil-temporal file
rnum: number of regions
coded by yufei tao, may 2003
******************************************************************/

void gen_qwrkld(char *_qfname, float _qs, int _qt, float _U, float _T, int _qnum, int _N,
				char *_dfname, char *_stfname, int _rnum)
{
	FILE *dfp=fopen(_dfname, "r");
	if (!dfp)
		fatal("could not open the spatial data file\n");
	FILE *rfp=fopen(_stfname, "r"); 
	if (!rfp)
		fatal("could not open the raw data file\n");
	FILE *qfp=fopen(_qfname, "w");
	if (!qfp)
		fatal("could not create the query file\n");

	printf("reading the spatial file...\n");
	float *reg=new float[4*_rnum];
	int cnt=0;
	while (!feof(dfp))
	{
		if (cnt==_rnum)
			error("\tthe spatial dataset's cardinality is larger than you thought...\n", true);
		int id; 
		fscanf(dfp, "%d ", &id);
		if (id!=cnt)
			error("the spatial file is not sorted according to id\n", true);
		for (int i=0; i<4; i++)
			fscanf(dfp, "%f ", &reg[4*cnt+i]);
		fscanf(dfp, "\n");
		cnt++;
	}
	printf("\tread %d regions\n", cnt);
	if (cnt!=_rnum)
		printf("\twarning: the spatial dataset's cardinality is smaller than you thought...\n");

	
	printf("generating queries...\n");
	float *qmbr=new float[_qnum*6];
	int *res=new int[_qnum];
	for (int i=0; i<_qnum; i++)
	{
		int regid=uniform(0, _rnum);
		if (regid==_rnum)
			regid--;

		qmbr[i*6]=reg[4*regid];
		qmbr[i*6+1]=qmbr[i*6]+_qs;
		qmbr[i*6+2]=reg[4*regid+2];
		qmbr[i*6+3]=qmbr[i*6+2]+_qs;
		qmbr[i*6+4]=(int)uniform(1, _T-_qt+2);
		if (qmbr[i*6+4]==_T-_qt+2)
			qmbr[i*6+4]--;
		qmbr[i*6+5]=qmbr[i*6+4]+_qt-1;

		res[i]=0;
	}

	printf("attaching the actual results to the queries...\n");
	bool *eids=new bool[_N*_qnum];
	for (i=0; i<_N*_qnum; i++)
		eids[i]=false;

	while (!feof(rfp)) 
	{
		int tm,rid,eid, year;
		fscanf(rfp, "%d %d %d %d", &tm, &rid, &eid, &year);

		for (int i=0; i<_qnum; i++)
		{
			float *tmp=overlapRect(2, &qmbr[6*i], &reg[rid*4]);
			if (tmp)
			{
				delete []tmp;
				if (qmbr[6*i+4]<=tm && tm<=qmbr[6*i+5])
				{
					if (!eids[_N*i+eid])
					{
						eids[_N*i+eid]=true; 
						res[i]++;
					}
				}
			}
		}
	}

	for (i=0; i<_qnum; i++)
	{
		for (int j=0; j<4; j++)
			fprintf(qfp, "%f ", qmbr[6*i+j]);
		fprintf(qfp, "%f %f %d\n", qmbr[6*i+4], qmbr[6*i+5], res[i]);
	}

	delete []eids;
	delete []reg;
	delete []res;
	delete []qmbr;
	fclose(qfp);
	fclose(rfp);
	fclose(dfp);	
}

/*****************************************************************
this function samples from a given spatial dataset
para:
dfname: the file name of the given dataset
tfname: the target file name
N: number of data regions
sN: the number of sample pts
coded by yufei tao, june 2003
******************************************************************/

void sample_ds(char *_dfname, char *_tfname, int _N, int _sN)
{
	FILE *dfp=fopen(_dfname, "r");
	if (!dfp)
		fatal("could not open the spatial data file\n");
	FILE *ofp=fopen(_tfname, "w");
	if (!ofp)
		fatal("could not create the target file name\n");

	printf("reading the spatial file...\n");
	float *reg=new float[4*_N];
	int cnt=0;
	while (!feof(dfp))
	{
		if (cnt==_N)
			error("\tthe spatial dataset's cardinality is larger than you thought...\n", true);
		int id; 
		fscanf(dfp, "%d ", &id);
		if (id!=cnt)
			error("the spatial file is not sorted according to id\n", true);
		for (int i=0; i<4; i++)
			fscanf(dfp, "%f ", &reg[4*cnt+i]);
		fscanf(dfp, "\n");
		cnt++;
	}
	printf("\tread %d regions\n", cnt);
	if (cnt!=_N)
		printf("\twarning: the spatial dataset's cardinality is smaller than you thought...\n");

	printf("now sampling...\n");
	int N=_N; //number of remaining regions
	for (int i=0; i<_sN; i++)
	{
		int sub=new_uniform(0, N); 
		if (sub==N) sub--;
		fprintf(ofp, "%d", i);
		for (int j=0; j<4; j++)
			fprintf(ofp, " %f", reg[4*sub+j]); 
		fprintf(ofp, "\n");
		//now move the subsequent elements one step forward
		for (j=sub; j<=N-2; j++)
			memcpy(&reg[4*sub], &reg[4*sub+4], sizeof(float)*4);
	}

	delete []reg;
	fclose(ofp);
	fclose(dfp);
}

void main()
{
	srand( (unsigned)time( NULL ) );

	float s=5;

	if (s==-1) gen_simple_ds("../ds/u1k.txt", 2, 10000, 1000, 1);
	if (s==0) normalize("../../real ds/lb.txt", "../../real ds/lb-1.txt", 2, 10000, 1);
	if (s==0.5)
		sample_ds("../../real ds/lb.txt", "../ds/lb10k.txt", 53145, 10000);
	if (s==1)
	{
		Cache *c=new Cache(0, 1024);
		RTree *rt=new RTree("../ds/lb.txt", "./trees/lb", 1024, c, 2);
		//RTree *rt=new RTree("../ds/u1k.txt", "./trees/u1k", 1024, c, 2);
		delete rt;
		delete c;
	}
	if (s==2) 
	{
		printf("reall generate moving ds? (y/n)");
		char c=getchar(); getchar();
		if (c!='y')
 			exit(1);
		//gen_moving_ds("../ds/lb.txt", "./trees/lb", "../ds/lb10k.st", 10000, 10000, 100, 1000, 2000);
		gen_moving_ds("../ds/lb.txt", "./trees/lb", "../ds/lb100k.st", 10000, 100000, 100, 1000, 2000);
	}
	if (s==2.5) //for datasets generated from gen_moving_ds, please sort them using the following function
	{
		ext_sort("../ds/lb100k.st", "../ds/lb100k.sts", 100000);
	}
	if (s==3) 
	{
		//gen_signatures("../ds/lb10k.sts", "../ds/lb10k.sig32-24", 32, 24, true, false);
		//gen_signatures("../ds/lb10k.sts", "../ds/lb10k.sig16-24", 16, 24, true, false);
		//gen_signatures("../ds/lb10k.sts", "../ds/lb10k.sig8-24", 8, 24, true, false);
		gen_signatures("../ds/lb100k.sts", "../ds/lb10k.sig16-24", 16, 24, true, false);
	}
	if (s==4)
	{
		int qrid[]={10000, 15000}; float qt[]={10,19};
		float qmbr[]={6048.653320, 7048.653320, 6251.960938, 7251.960938};
		//test_accuracy("../ds/lb.txt", "../ds/lb10k.sig32-24", "../ds/lb10k.sts", qmbr, qt, 
		//	10000, 10000, 32, 24, true);
		test_accuracy("../ds/lb.txt", "../ds/lb10k.sig16-24", "../ds/lb10k.sts", qmbr, qt, 
			10000, 10000, 16, 24, true);
	}
	if (s==5)
	{
	//	gen_qwrkld("../qs/lb10k.1k-10", 1000, 10, 10000, 100, 50, 10000, "../ds/lb.txt", 
	//	"../ds/lb10k.sts", 10000);
//		gen_qwrkld("../qs/lb10k.1.5k-10", 1500, 10, 10000, 100, 50, 10000, "../ds/lb.txt", 
//		"../ds/lb10k.sts", 10000);
//		gen_qwrkld("../qs/lb10k.2k-10", 2000, 10, 10000, 100, 50, 10000, "../ds/lb.txt", 
//		"../ds/lb10k.sts", 10000);
//		gen_qwrkld("../qs/lb10k.0.5k-10", 500, 10, 10000, 100, 50, 10000, "../ds/lb.txt", 
//		"../ds/lb10k.sts", 10000);
//		gen_qwrkld("../qs/lb10k.2.5k-10", 2500, 10, 10000, 100, 50, 10000, "../ds/lb.txt", 
//		"../ds/lb10k.sts", 10000);
//		gen_qwrkld("../qs/lb10k.1.5k-1", 1500, 1, 10000, 100, 50, 10000, "../ds/lb.txt", 
//		"../ds/lb10k.sts", 10000);
//		gen_qwrkld("../qs/lb10k.1.5k-20", 1500, 20, 10000, 100, 50, 10000, "../ds/lb.txt", 
//		"../ds/lb10k.sts", 10000);
		gen_qwrkld("../qs/lb100k.1.5k-10", 1500, 10, 10000, 100, 50, 100000, "../ds/lb.txt", 
		"../ds/lb100k.sts", 10000);
	}
}